@@ -20,12 +20,13 @@ use super::term::*;
 use crate::{term_getable, term_putable, unifiable};
 use std::convert::TryInto;
 use std::os::raw::c_char;
+use std::sync::atomic::{AtomicUsize, Ordering};
 
 /// A wrapper for a prolog atom.
 ///
 /// When created, the underlying atom will have its reference count
 /// increased. When dropped, the reference count will decrease.
-#[derive(PartialEq, Eq, Debug)]
+#[derive(PartialEq, Eq, Hash, Debug)]
 pub struct Atom {
     atom: atom_t,
 }
@@ -122,6 +123,11 @@ impl Atom {
 
         swipl_string
     }
+
+    /// Increase the reference counter for this atom.
+    fn increment_refcount(&self) {
+        unsafe { PL_register_atom(self.atom) }
+    }
 }
 
 impl Clone for Atom {
@@ -161,7 +167,7 @@ where
     let mut atom = 0;
     let result = unsafe { PL_get_atom(term.term_ptr(), &mut atom) };
 
-    if unsafe { PL_exception(0) != 0 } {
+    if unsafe { pl_default_exception() != 0 } {
         return Err(PrologError::Exception);
     }
 
@@ -285,12 +291,28 @@ pub trait AsAtom {
     /// The second argument ensures that this function is called in a
     /// context where atoms are allowed to be created.
     fn as_atom(&self) -> Atom;
+
+    /// Turn the borrowed object into an `atom_t`, and returns an
+    /// allocation which will keep this `atom_t` valid as long as it
+    /// is not dropped.
+    ///
+    /// This allows code that takes an `AsAtom` to be a little bit
+    /// smart about not cloning the underlying data, if the underlying
+    /// data is already an atom.
+    fn as_atom_ptr(&self) -> (atom_t, Option<Atom>) {
+        let atom = self.as_atom();
+        (atom.atom_ptr(), Some(atom))
+    }
 }
 
 impl AsAtom for Atom {
     fn as_atom(&self) -> Atom {
         self.clone()
     }
+
+    fn as_atom_ptr(&self) -> (atom_t, Option<Atom>) {
+        (self.atom_ptr(), None)
+    }
 }
 
 impl<'a> AsAtom for Atomable<'a> {
@@ -299,6 +321,12 @@ impl<'a> AsAtom for Atomable<'a> {
     }
 }
 
+impl<'a> AsAtom for &'a str {
+    fn as_atom(&self) -> Atom {
+        self.into_atom()
+    }
+}
+
 impl AsAtom for str {
     fn as_atom(&self) -> Atom {
         self.into_atom()
@@ -343,7 +371,7 @@ where
         )
     };
 
-    if unsafe { PL_exception(0) != 0 } {
+    if unsafe { pl_default_exception() != 0 } {
         return Err(PrologError::Exception);
     }
 
@@ -395,6 +423,63 @@ impl<'a> AsRef<str> for Atomable<'a> {
     }
 }
 
+/// A struct which provides a way to delay and cache atom creation.
+///
+/// This struct wraps a static string and uses it to construct a swipl
+/// atom on the first invocation of `as_atom`. Subsequent invocations
+/// will reuse the earlier constructed atom.
+///
+/// The purpose of this struct is to back the implementation of the
+/// `atom!` macro, but it's also usable on its own.
+pub struct LazyAtom {
+    s: &'static str,
+    a: AtomicUsize,
+}
+
+impl LazyAtom {
+    /// Create a new LazyAtom.
+    ///
+    /// This constructor is const, as all it does is set up the
+    /// struct. No actual calls into SWI-Prolog happen at this stage.
+    pub const fn new(s: &'static str) -> Self {
+        Self {
+            s,
+            a: AtomicUsize::new(0),
+        }
+    }
+
+    /// Create an atom, or return an earlier created atom.
+    ///
+    /// On first call, this will call swipl to create an
+    /// atom. Subsequent calls will reuse the atom that was retrieved
+    /// before.
+    pub fn as_atom(&self) -> Atom {
+        let ptr = self.a.load(Ordering::Relaxed);
+        if ptr == 0 {
+            // we've not yet allocated an atom. let's do it now.
+            let atom = Atom::new(self.s);
+
+            let swapped = self.a.swap(atom.atom_ptr() as atom_t, Ordering::Relaxed);
+            if swapped == 0 {
+                // nobody raced us to store this atom. This means
+                // we're the first ones here. We need to ensure that
+                // the atom refcount is incremented so that our
+                // reference here in static memory remains valid.
+
+                atom.increment_refcount();
+            }
+
+            atom
+        } else {
+            let atom = unsafe { Atom::wrap(ptr as atom_t) };
+
+            atom.increment_refcount();
+
+            atom
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -547,4 +632,39 @@ mod tests {
 
         assert_eq!("foo", a2.name());
     }
+
+    #[test]
+    fn lazy_atom_to_atom() {
+        let engine = Engine::new();
+        let _activation = engine.activate();
+
+        let lazy = LazyAtom::new("moo");
+        let a1 = lazy.as_atom();
+        let a2 = lazy.as_atom();
+
+        assert_eq!(a1, a2);
+        let a3 = "moo".as_atom();
+        assert_eq!(a1, a3);
+    }
+
+    use swipl_macros::atom;
+    #[test]
+    fn inline_atom_through_macro_ident() {
+        let engine = Engine::new();
+        let _activation = engine.activate();
+
+        let a1 = atom!(foo);
+        let a2 = "foo".as_atom();
+        assert_eq!(a1, a2);
+    }
+
+    #[test]
+    fn inline_atom_through_macro_str() {
+        let engine = Engine::new();
+        let _activation = engine.activate();
+
+        let a1 = atom!("bar");
+        let a2 = "bar".as_atom();
+        assert_eq!(a1, a2);
+    }
 }
