@@ -38,7 +38,7 @@ impl<const N: usize> LazyCallablePredicate<N> {
     /// what it previously looked up. Otherwise, it'll do the lookup.
     pub fn as_callable(&self) -> CallablePredicate<N> {
         assert_some_engine_is_active();
-        let mut loaded = self.predicate.load(Ordering::Relaxed);
+        let mut loaded: predicate_t = self.predicate.load(Ordering::Relaxed) as predicate_t;
         if loaded.is_null() {
             let functor = Functor::new(self.name, N as u16);
             let module_name = self.module.unwrap_or("");
@@ -47,7 +47,7 @@ impl<const N: usize> LazyCallablePredicate<N> {
             loaded = Predicate::new(functor, module).predicate_ptr();
 
             self.predicate
-                .store(loaded, std::sync::atomic::Ordering::Relaxed);
+                .store(loaded as *mut c_void, std::sync::atomic::Ordering::Relaxed);
         }
 
         unsafe { CallablePredicate::wrap(loaded) }
@@ -320,7 +320,7 @@ impl<const N: usize> Callable<N> for CallablePredicate<N> {
 mod tests {
     use crate::prelude::*;
     #[test]
-    fn call_prolog_inline() -> PrologResult<()> {
+    fn call_prolog_inline_slash() -> PrologResult<()> {
         let engine = Engine::new();
         let activation = engine.activate();
         let context: Context<_> = activation.into();
@@ -331,4 +331,56 @@ mod tests {
 
         Ok(())
     }
+
+    #[test]
+    fn call_prolog_inline_module_slash() -> PrologResult<()> {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term = term! {context: flurps(flargh)}?;
+        context.call_once(pred!(user: writeq / 1), [&term]).unwrap();
+        context.call_once(pred!(user: nl / 0), []).unwrap();
+
+        Ok(())
+    }
+
+    #[test]
+    fn call_prolog_inline_str() -> PrologResult<()> {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term = term! {context: flurps(flargh)}?;
+        context.call_once(pred!("writeq/1"), [&term]).unwrap();
+        context.call_once(pred!("nl/0"), []).unwrap();
+
+        Ok(())
+    }
+
+    #[test]
+    fn call_prolog_inline_module_str() -> PrologResult<()> {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term = term! {context: flurps(flargh)}?;
+        context.call_once(pred!(user:"writeq/1"), [&term]).unwrap();
+        context.call_once(pred!(user:"nl/0"), []).unwrap();
+
+        Ok(())
+    }
+
+    #[test]
+    fn call_prolog_inline_str_with_module() -> PrologResult<()> {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term = term! {context: flurps(flargh)}?;
+        context.call_once(pred!("user:writeq/1"), [&term]).unwrap();
+        context.call_once(pred!("user:nl/0"), []).unwrap();
+
+        Ok(())
+    }
 }
