@@ -16,6 +16,7 @@ pub mod atom;
 pub mod blob;
 pub mod callable;
 pub mod context;
+pub mod dict;
 pub mod engine;
 pub mod functor;
 pub mod init;
@@ -29,5 +30,6 @@ pub mod text;
 pub mod prelude;
 
 pub use swipl_macros::{
-    arc_blob, clone_blob, pred, predicates, prolog, term, wrapped_arc_blob, wrapped_clone_blob,
+    arc_blob, atom, clone_blob, functor, pred, predicates, prolog, term, wrapped_arc_blob,
+    wrapped_clone_blob,
 };
