@@ -34,6 +34,7 @@ use super::context::*;
 use super::engine::*;
 use super::fli::*;
 use super::result::*;
+use std::cmp::{Ordering, PartialOrd};
 use std::convert::TryInto;
 use std::fmt;
 use std::fmt::Debug;
@@ -64,12 +65,34 @@ impl<'a> Term<'a> {
         self.term
     }
 
+    pub fn origin_engine_ptr(&self) -> PL_engine_t {
+        self.origin.origin_engine_ptr()
+    }
+
     /// Returns true if this term reference holds a variable.
     pub fn is_var(&self) -> bool {
         self.assert_term_handling_possible();
         unsafe { PL_is_variable(self.term) != 0 }
     }
 
+    /// Returns true if this term reference holds an atom.
+    pub fn is_atom(&self) -> bool {
+        self.assert_term_handling_possible();
+        unsafe { PL_is_atom(self.term) != 0 }
+    }
+
+    /// Returns true if this term reference holds a string.
+    pub fn is_string(&self) -> bool {
+        self.assert_term_handling_possible();
+        unsafe { PL_is_string(self.term) != 0 }
+    }
+
+    /// Returns true if this term reference holds an integer.
+    pub fn is_integer(&self) -> bool {
+        self.assert_term_handling_possible();
+        unsafe { PL_is_integer(self.term) != 0 }
+    }
+
     /// Reset terms created after this term, including this term itself.
     /// Only safe to call when you're sure these terms aren't used afterwards.
     pub unsafe fn reset(&self) {
@@ -101,7 +124,7 @@ impl<'a> Term<'a> {
         // through some inner function. It's too much hassle to have
         // each function check for themselves and return the
         // appropriate thing, so we just do it all here.
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             Err(PrologError::Exception)
         } else if result {
             Ok(())
@@ -131,7 +154,7 @@ impl<'a> Term<'a> {
         let arg_ref = unsafe { PL_new_term_ref() };
 
         let result = unsafe { PL_get_arg(index.try_into().unwrap(), self.term, arg_ref) };
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             return Err(PrologError::Exception);
         }
 
@@ -164,7 +187,7 @@ impl<'a> Term<'a> {
         // through some inner function. It's too much hassle to have
         // each function check for themselves and return the
         // appropriate thing, so we just do it all here.
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             Err(PrologError::Exception)
         } else if opt.is_none() {
             Err(PrologError::Failure)
@@ -204,7 +227,7 @@ impl<'a> Term<'a> {
         // through some inner function. It's too much hassle to have
         // each function check for themselves and return the
         // appropriate thing, so we just do it all here.
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             Err(PrologError::Exception)
         } else if opt.is_none() {
             let context = unsafe { unmanaged_engine_context() };
@@ -242,7 +265,7 @@ impl<'a> Term<'a> {
         let arg_ref = unsafe { PL_new_term_ref() };
 
         let result = unsafe { PL_get_arg(index.try_into().unwrap(), self.term, arg_ref) };
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             return Err(PrologError::Exception);
         }
 
@@ -277,7 +300,7 @@ impl<'a> Term<'a> {
             )
         };
 
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             return Err(PrologError::Exception);
         }
 
@@ -327,7 +350,7 @@ impl<'a> Term<'a> {
             )
         };
 
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             return Err(PrologError::Exception);
         }
 
@@ -360,7 +383,7 @@ impl<'a> Term<'a> {
     pub fn put<T: TermPutable + ?Sized>(&self, val: &T) -> NonFailingPrologResult<()> {
         val.put(self);
 
-        if unsafe { PL_exception(0) != 0 } {
+        if unsafe { pl_default_exception() != 0 } {
             Err(PrologException)
         } else {
             Ok(())
@@ -384,6 +407,45 @@ impl<'a> Term<'a> {
     }
 }
 
+impl<'a> PartialEq for Term<'a> {
+    fn eq(&self, other: &Term) -> bool {
+        self.assert_term_handling_possible();
+        if self.origin.origin_engine_ptr() != other.origin.origin_engine_ptr() {
+            panic!("terms being compard are not part of the same engine");
+        }
+
+        let result = unsafe { PL_compare(self.term_ptr(), other.term_ptr()) };
+
+        result == 0
+    }
+}
+impl<'a> PartialOrd for Term<'a> {
+    fn partial_cmp(&self, other: &Term) -> Option<Ordering> {
+        self.assert_term_handling_possible();
+        if self.origin.origin_engine_ptr() != other.origin.origin_engine_ptr() {
+            panic!("terms being compard are not part of the same engine");
+        }
+
+        let result = unsafe { PL_compare(self.term_ptr(), other.term_ptr()) };
+
+        Some(if result < 0 {
+            Ordering::Less
+        } else if result == 0 {
+            Ordering::Equal
+        } else {
+            Ordering::Greater
+        })
+    }
+}
+
+impl<'a> Eq for Term<'a> {}
+
+impl<'a> Ord for Term<'a> {
+    fn cmp(&self, other: &Term) -> Ordering {
+        self.partial_cmp(other).unwrap()
+    }
+}
+
 #[derive(Clone)]
 pub(crate) struct TermOrigin<'a> {
     engine: PL_engine_t,
@@ -687,7 +749,7 @@ unifiable! {
         // compatibility functions work.
         unsafe {with_cleared_exception(|| {
             let result = PL_unify_uint64(term.term, *self);
-            let error_term_ref = PL_exception(0);
+            let error_term_ref = pl_default_exception();
             if error_term_ref != 0 {
                 let ctx = unmanaged_engine_context();
                 let error_term = ctx.new_term_ref();
@@ -723,7 +785,7 @@ term_getable! {
         unsafe {with_cleared_exception(|| {
             let mut out = 0;
             let result = PL_cvt_i_uint64(term.term, &mut out);
-            let error_term_ref = PL_exception(0);
+            let error_term_ref = pl_default_exception();
             if error_term_ref != 0 {
                 let ctx = unmanaged_engine_context();
                 let error_term = ctx.new_term_ref();
@@ -944,25 +1006,26 @@ where
 
         let frame = context.open_frame();
         let list = frame.new_term_ref();
+
         if list.unify(term).is_err() {
             return false;
         }
-        let mut success = true;
 
         for t in self.iter() {
             // create a new frame to ensure we don't just keep putting head and tail refs on the stack.
             let frame2 = frame.open_frame();
             let head = frame2.new_term_ref();
             let tail = frame2.new_term_ref();
-            success =
-                unsafe { PL_unify_list(list.term_ptr(), head.term_ptr(), tail.term_ptr()) != 0 };
 
-            if !success {
-                break;
+            // if list unification fails, or head can not be unified with current term,
+            // return false early.
+            // note: || is short-circuiting OR
+            if unsafe { PL_unify_list(list.term_ptr(), head.term_ptr(), tail.term_ptr()) == 0 }
+                || head.unify(t).is_err()
+            {
+                return false;
             }
 
-            success = head.unify(t).is_ok();
-
             // reset term - should really be a method on term
             unsafe { PL_put_variable(list.term_ptr()) };
 
@@ -973,10 +1036,8 @@ where
             frame2.close();
         }
 
-        if success {
-            success = unsafe { PL_unify_nil(list.term_ptr()) != 0 };
-            frame.close();
-        }
+        let success = unsafe { PL_unify_nil(list.term_ptr()) != 0 };
+        frame.close();
 
         success
     }
@@ -1040,9 +1101,7 @@ unsafe impl<T: TermGetable> TermGetable for Vec<T> {
 
 #[cfg(test)]
 mod tests {
-    use crate::context::*;
-    use crate::engine::*;
-    use crate::result::*;
+    use super::*;
 
     #[test]
     fn unify_some_terms_with_success() {
@@ -1251,4 +1310,66 @@ mod tests {
 
         term.unify(binary_data.as_slice())
     }
+
+    #[test]
+    fn unify_list_should_fail_for_early_unmatched_value() {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term = context.new_term_ref();
+        term.unify([1u64, 2u64, 3u64].as_slice()).unwrap();
+
+        assert!(!attempt(term.unify([2u64, 2u64, 3u64].as_slice())).unwrap());
+    }
+
+    #[test]
+    fn term_equality() {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let [term1, term2, term3] = context.new_term_refs();
+        term1.unify(42_u64).unwrap();
+        term2.unify(42_u64).unwrap();
+        term3.unify(43_u64).unwrap();
+        assert_eq!(term1, term2);
+        assert_ne!(term1, term3);
+    }
+
+    #[test]
+    fn term_equality_different_context() {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let term1 = context.new_term_ref();
+        let frame = context.open_frame();
+        let [term2, term3] = frame.new_term_refs();
+
+        term1.unify(42_u64).unwrap();
+        term2.unify(42_u64).unwrap();
+        term3.unify(43_u64).unwrap();
+        assert_eq!(term1, term2);
+        assert_ne!(term1, term3);
+    }
+
+    #[test]
+    fn term_order() {
+        let engine = Engine::new();
+        let activation = engine.activate();
+        let context: Context<_> = activation.into();
+
+        let mut terms: [Term; 3] = context.new_term_refs();
+
+        terms[0].unify("foo").unwrap();
+        terms[1].unify("bar").unwrap();
+        terms[2].unify("baz").unwrap();
+
+        terms.sort();
+
+        assert_eq!("bar", terms[0].get::<String>().unwrap());
+        assert_eq!("baz", terms[1].get::<String>().unwrap());
+        assert_eq!("foo", terms[2].get::<String>().unwrap());
+    }
 }
